"""Temporary directory utilities

"""
import tempfile


class TemporaryDirectory(tempfile.TemporaryDirectory):
    """
    like tempfile.TemporaryDirectory but with an optional argument to not
    clean the directory afterwards.

    """

    def __init__(self, suffix=None, prefix=None, dir=None, clean_dir=True):
        self.clean_dir = clean_dir
        if self.clean_dir:
            super(TemporaryDirectory, self).__init__(suffix=suffix, prefix=prefix, dir=dir)
        else:
            self.name = tempfile.mkdtemp(suffix, prefix, dir)

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.clean_dir:
            super(TemporaryDirectory, self).__exit__(exc_type, exc_val, exc_tb)




"""JSON encoding and decoding tools for RAMP objects"""
import json
from typing import Any, Hashable

from ramp_core.serializable import Serializable

IDENTIFIER = 'cls'
DATA = '__data'


class RampJSONEncoder(json.JSONEncoder):
    """JSONEncoder for ramp objects that can handle Serializable objects.

    It is useful to have a single json-compatible class to handle our long-term serialization of Ramp objects.

    """

    def default(self, obj):
        try:
            ident, s = obj.serialize()
            return s | {IDENTIFIER: ident} if isinstance(s, dict) else {IDENTIFIER: ident, DATA: s}
        except AttributeError:
            pass
        return json.JSONEncoder.default(self, obj)


class RampJSONDecoder(json.JSONDecoder):
    """JSONDecoder for ramp objects that can reconstruct Serializable objects.

    It will be useful to be able to recreate our Python objects for the ramp objects if we serialize them in a
    long term, non-pickle method.

    The idea is that we save an identifier which allows us to know which Python object to reconstruct.
    To avoid the Pickle problem, the way we do this is that we assume the decoding user can create a mapping of
    serialization identifiers to factory classes that have a deserialize method.

    Examples
    --------
    >>> class A:
    ...     ser_identifier = "AA"
    ...
    ...     def __init__(self, a):
    ...         self.a = a
    ...
    ...     def serialize(self):
    ...         return self.ser_identifier, {'a': self.a}
    ...
    ...     @classmethod
    ...     def deserialize(cls, d, **_):
    ...         return cls(**d)
    >>>
    >>> b = A(5)
    >>> s = json.dumps(b, cls=RampJSONEncoder)
    >>> try:  # This fails because we didn't set the supported attribute
    ...     v = json.loads(s, cls=RampJSONDecoder)
    ... except AttributeError:
    ...     pass
    ... else:
    ...     raise RuntimeError("Should have crashed")
    >>> RampJSONDecoder.supported = {type(b).ser_identifier: b}
    >>> v = json.loads(s, cls=RampJSONDecoder)  # And now it succeeds
    >>> v.a == b.a
    True

    """

    supported: dict[str, Serializable]


    def decode(self, s: str, **kw):
        pobj = super().decode(s, **kw)
        if not isinstance(pobj, dict):
            return pobj
        if IDENTIFIER not in pobj:
            return pobj
        styp = pobj[IDENTIFIER]
        if styp in self.supported:
            del pobj[IDENTIFIER]
            typ = self.supported[styp]
            if DATA in pobj:
                return typ.deserialize(pobj[DATA], supported=self.supported)
            return typ.deserialize(pobj, supported=self.supported)
        return pobj


def unserializable(d: dict | list | tuple) -> list[tuple[list[Hashable], Any]]:
    """Returns which items were unserializable. Useful to debug serializability.
    """
    lst = []
    if isinstance(d, dict):
        iters = d.items()
    elif isinstance(d, (list, tuple)):
        iters = enumerate(d)
    else:
        try:
            json.dumps(d, cls=RampJSONEncoder)
        except TypeError:
            return [([], d)]
        else:
            return []

    for a, b in iters:
        if isinstance(b, (dict, list, tuple)):
            lst.extend([([a] + access, v) for access, v in unserializable(b)])
        else:
            try:
                json.dumps(b, cls=RampJSONEncoder)
            except TypeError:
                lst.append(([a], b))
    return lst




from string import ascii_lowercase

import hypothesis.strategies as st
from hypothesis import given

from ramp_core.serializable import _prefer_list


@given(st.text(alphabet=ascii_lowercase, min_size=1))
def test_prefer_list_for_str_is_str(sold):
    s = _prefer_list(sold)
    assert type(s) is str
    assert s == sold


@given(st.lists(elements=st.integers(), min_size=1).map(tuple))
def test_prefer_list_for_tuple_is_list(x):
    y = _prefer_list(x)
    assert type(y) is list
    assert all(yv == xv for yv, xv in zip(y, x))



"""Defines the Serializable protocol, which is what we expect things to have, so we can serialize and deserialize them

"""

from pathlib import PurePath
from typing import Any, ClassVar, Protocol, Sequence, Type, TypeVar, runtime_checkable

try:  # To allow for older Python versions where Self was not supported
    from typing import Self
except ImportError:
    Self = TypeVar("Self")

import numpy as np
from cytoolz import valmap

T = TypeVar("T", bound="Serializable")
D = TypeVar("D", bound="_Deserializable")


class _Deserializable(Protocol):
    @classmethod
    def deserialize(cls: Type[D], d: dict[str, Any], *, supported: dict[str, Type["Serializable"]]) -> D:
        ...


@runtime_checkable
class Serializable(Protocol):
    """Protocol for classes that we can serialize to JSON-able formats and deserialize from.

    Serializable objects must be capable of being equal to themselves by going through serialization and deserialization

    """

    ser_identifier: ClassVar[str]

    def serialize(self) -> tuple[str, dict[str, Any]]:
        """Serializes the object to return its class' identifier and the data that allows to recreate the instance.

        Returns
        -------
        str, dict[str, data]
            The ser_identifier and the data that allows to recreate an instance.

        """
        return self.ser_identifier, _ensure_list(_save_attributes(self))

    @classmethod
    def deserialize(cls: Type[Self], d: dict[str, Any], *, supported: dict[str, Type["Serializable"]]) -> Self:
        """Create a new instance given the data generated by an instance's serialize method.

        Parameters
        ----------
        d: dict[str, Any]
            The data dictionary returned by the original object's serialize method.
        supported: dict[str, Type[Serializable]]
            A dictionary of which types are supported for deserialization.

        """
        return cls(**_default_tuple(d))

    def __eq__(self: T, other: T) -> bool:
        ...


def deserialize_default(data: tuple[str, dict[str, Any]],
                        supported: dict[str, Type[Serializable]],
                        default: Type[_Deserializable] | None = None) -> _Deserializable:
    """Deserialize the data given by a Serializable object's `serialize` method given class deserialization

    Parameters
    ----------
    data: str, dict[str, Any]
        The result from the object's Serialize method. See :func:Serializable.serialize.
    supported: dict[str, Type[Serializable]]
        A mapping of serialization identifiers to the classes they represent.
    default: Type[Serializable] or None
        The default type to use if the serialization identifier is not recognized. If None, raises an error if the
        identifier isn't there.

    Returns
    -------
    Serializable
        The deserialized object

    """
    name, d = data
    if default is not None:
        typ = supported.get(name, default)
    else:
        if name not in supported:
            raise TypeError(f"Deserialization of {name} from json isn't supported by the deserializer")
        typ = supported[name]
    return typ.deserialize(d, supported=supported)


def _save_attributes(self) -> dict:
    sd = {k: self.__getattribute__(k) for k in self.__slots__} if self.__slots__ else {}
    return self.__dict__ | sd


def _default_tuple(d: dict) -> dict:
    return valmap(lambda x: tuple(x) if isinstance(x, list) else x, d)


def _prefer_list(x):
    if isinstance(x, Sequence) and not isinstance(x, str):
        return list(x)
    if isinstance(x, np.ndarray):
        return x.tolist()
    if isinstance(x, PurePath):
        return str(x)
    return x


def _ensure_list(d: dict) -> dict:
    return valmap(_prefer_list, d)
